package it.unimi.dsi.webgraph.algo;

/*		 
 * Copyright (C) 2010-2011 Paolo Boldi & Sebastiano Vigna 
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 3 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 */


import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import it.unimi.dsi.fastutil.longs.LongBigList;
import it.unimi.dsi.util.IntHyperLogLogCounterArray;
import it.unimi.dsi.webgraph.ArrayListMutableGraph;
import it.unimi.dsi.webgraph.ImmutableGraph;
import it.unimi.dsi.webgraph.Transform;
import it.unimi.dsi.webgraph.WebGraphTestCase;
import it.unimi.dsi.webgraph.examples.ErdosRenyiGraph;

import java.io.IOException;

import org.junit.Test;


public class HyperApproximateNeighbourhoodFunctionTest extends WebGraphTestCase {
	// Below this threshold errors due to block-by-block summing start to appear.
	protected static final double THRESHOLD = 1E-9;

	/** Checks that the state of two HyperApproximateNeighbourhoodFunction implementation (as
	 * returned by {@link IntHyperLogLogCounterArray#registers()}) are exactly the same. */
	public final static void assertState( final int size, final int log2m, final LongBigList[] a, final LongBigList[] b ) {
		final int m = 1 << log2m;
		for( int i = 0; i < size; i++ ) {
			for( int j = 0; j < m; j++ ) {
				final long index = ( (long)i << log2m ) + j;
				final int chunk = (int)( index >>> IntHyperLogLogCounterArray.CHUNK_SHIFT ); 
				final long offset = index & IntHyperLogLogCounterArray.CHUNK_MASK;
				assertEquals( "Counter " + i + ", register " + j + ": ", a[ chunk ].getLong( offset ), b[ chunk ].getLong( offset ) );
			}
		}
	}

	@Test
	public void testTrivial() throws IOException {
		ImmutableGraph g = ArrayListMutableGraph.newCompleteBinaryIntree( 10 ).immutableView();
		HyperApproximateNeighbourhoodFunction hyperApproximateNeighbourhoodFunction = new HyperApproximateNeighbourhoodFunction( g, g, 7, null, 0, 0, 0, false, false, false, 0 );
		hyperApproximateNeighbourhoodFunction.approximateNeighbourhoodFunction( Long.MAX_VALUE, -1 );
		hyperApproximateNeighbourhoodFunction.approximateNeighbourhoodFunction( Long.MAX_VALUE, -1 );

		hyperApproximateNeighbourhoodFunction = new HyperApproximateNeighbourhoodFunction( g, g, 7, null, 0, 0, 0, true, false, false, 0 );
		hyperApproximateNeighbourhoodFunction.approximateNeighbourhoodFunction( Long.MAX_VALUE, -1 );
		hyperApproximateNeighbourhoodFunction.approximateNeighbourhoodFunction( Long.MAX_VALUE, -1 );
	}
	
	protected static void assertRelativeError( double sequentialCurrent, double current, double threshold ) {
		assertTrue( sequentialCurrent + " != " + current + ", " + Math.abs( current - sequentialCurrent ) / current + " > " + threshold, Math.abs( current - sequentialCurrent ) / current <= THRESHOLD );
	}
	
	/* All tests in this class check that 2 times the theoretical relative standard deviation
	 * is attained in 9 trials out of 10. The theory (in particular, the Vysochanskii-Petunin inequality) 
	 * indeed says it should happen 90% of the times. */
	
	@Test
	public void testClique() throws IOException {
		for( int log2m: new int[] { 4, 5, 6, 8 } ) {
			final double rsd = HyperApproximateNeighbourhoodFunction.relativeStandardDeviation( log2m );
			for( int size: new int[] { 10, 100, 500 } ) {
				int correct = 0;
				for( int attempt = 0; attempt < 10; attempt++ ) {
					System.err.println( "log2m: " + log2m + " size: " + size + " attempt: " + attempt );
					ImmutableGraph g = ArrayListMutableGraph.newCompleteGraph( size, false ).immutableView();
					HyperApproximateNeighbourhoodFunction hyperApproximateNeighbourhoodFunction = new HyperApproximateNeighbourhoodFunction( g, attempt % 3 == 0 ? null : Transform.transpose( g ), log2m, null, 0, 10, 10, attempt % 2 == 0, false, false, attempt );
					SequentialHyperApproximateNeighbourhoodFunction sequentialHyperApproximateNeighbourhoodFunction = new SequentialHyperApproximateNeighbourhoodFunction( g, log2m, null, attempt );
					hyperApproximateNeighbourhoodFunction.init();
					sequentialHyperApproximateNeighbourhoodFunction.init();
					final double current = hyperApproximateNeighbourhoodFunction.iterate();
					final double sequentialCurrent = sequentialHyperApproximateNeighbourhoodFunction.iterate();

					assertState( size, log2m, sequentialHyperApproximateNeighbourhoodFunction.registers(), hyperApproximateNeighbourhoodFunction.registers() );

					if ( Math.abs( size * size - current ) <= 2 * rsd * size * size ) correct++;

					assertRelativeError( sequentialCurrent, current, THRESHOLD );
					
					hyperApproximateNeighbourhoodFunction.close();
					sequentialHyperApproximateNeighbourhoodFunction.close();
				}
				assertTrue( size + ":" + rsd + " " + correct + " < " + 9, correct >= 9 );
			}
		}
	}

	@Test
	public void testErdosRenyi() throws IOException {
		for( int log2m: new int[] { 4, 5, 6, 8 } ) {
			for( int size: new int[] { 10, 100, 500 } ) {
				for( int attempt = 0; attempt < 10; attempt++ ) {
					System.err.println( "log2m: " + log2m + " size: " + size + " attempt: " + attempt );
					ImmutableGraph g = new ArrayListMutableGraph( new ErdosRenyiGraph( size, .1, attempt, false ) ).immutableView();
					HyperApproximateNeighbourhoodFunction hyperApproximateNeighbourhoodFunction = new HyperApproximateNeighbourhoodFunction( g, attempt % 3 == 0 ? null : Transform.transpose( g ), log2m, null, 0, 10 * (attempt % 3), 10, attempt % 2 == 0, false, false, attempt );
					SequentialHyperApproximateNeighbourhoodFunction sequentialHyperApproximateNeighbourhoodFunction = new SequentialHyperApproximateNeighbourhoodFunction( g, log2m, null, attempt );
					hyperApproximateNeighbourhoodFunction.init();
					sequentialHyperApproximateNeighbourhoodFunction.init();
					do {
						final double current = hyperApproximateNeighbourhoodFunction.iterate();
						final double sequentialCurrent = sequentialHyperApproximateNeighbourhoodFunction.iterate();
						assertState( size, log2m, sequentialHyperApproximateNeighbourhoodFunction.registers(), hyperApproximateNeighbourhoodFunction.registers() );
						assertRelativeError( sequentialCurrent, current, THRESHOLD );
					} while( hyperApproximateNeighbourhoodFunction.modified() != 0 );

					hyperApproximateNeighbourhoodFunction.init();
					sequentialHyperApproximateNeighbourhoodFunction.init();
					do {
						final double current = hyperApproximateNeighbourhoodFunction.iterate();
						final double sequentialCurrent = sequentialHyperApproximateNeighbourhoodFunction.iterate();
						assertState( size, log2m, sequentialHyperApproximateNeighbourhoodFunction.registers(), hyperApproximateNeighbourhoodFunction.registers() );
						assertRelativeError( sequentialCurrent, current, THRESHOLD );
					} while( hyperApproximateNeighbourhoodFunction.modified() != 0 );

					hyperApproximateNeighbourhoodFunction.close();
					sequentialHyperApproximateNeighbourhoodFunction.close();
				}
			}
		}
	}

	@Test
	public void testCycle() throws IOException {
		for( int log2m: new int[] { 4, 5, 6 } ) {
			final double rsd = HyperApproximateNeighbourhoodFunction.relativeStandardDeviation( log2m );
			for( int size: new int[] { 100, 500, 1000 } ) {
				final int[] correct = new int[ size + 1 ];
				for( int attempt = 0; attempt < 10; attempt++ ) {
					System.err.println( "log2m: " + log2m + " size: " + size + " attempt: " + attempt );
					ImmutableGraph g = ArrayListMutableGraph.newDirectedCycle( size ).immutableView();
					HyperApproximateNeighbourhoodFunction hyperApproximateNeighbourhoodFunction = new HyperApproximateNeighbourhoodFunction( g, attempt % 3 == 0 ? null : Transform.transpose( g ), log2m, null, 0, 10 * (attempt % 3), 10, attempt % 2 == 0, false, false, attempt );
					SequentialHyperApproximateNeighbourhoodFunction sequentialHyperApproximateNeighbourhoodFunction = new SequentialHyperApproximateNeighbourhoodFunction( g, log2m, null, attempt );
					hyperApproximateNeighbourhoodFunction.init();
					sequentialHyperApproximateNeighbourhoodFunction.init();
					for( int i = 2; i <= size; i++ ) {
						final double current = hyperApproximateNeighbourhoodFunction.iterate();
						final double sequentialCurrent = sequentialHyperApproximateNeighbourhoodFunction.iterate();
						
						assertState( size, log2m, sequentialHyperApproximateNeighbourhoodFunction.registers(), hyperApproximateNeighbourhoodFunction.registers() );
						assertRelativeError( sequentialCurrent, current, THRESHOLD );
						if ( Math.abs( size * i - current ) <= 2 * rsd * size * i ) correct[ i ]++;
					}
					hyperApproximateNeighbourhoodFunction.close();
					sequentialHyperApproximateNeighbourhoodFunction.close();
				}
				for( int i = 2; i <= size; i++ ) assertTrue( size + ":" + rsd + " " + correct[ i ] + " < " + 9, correct[ i ] >= 9 );
			}
		}

	}

	@Test
	public void testOutdirectedStar() throws IOException {
		for( int log2m: new int[] { 4, 5, 6 } ) {
			final double rsd = HyperApproximateNeighbourhoodFunction.relativeStandardDeviation( log2m );
			for( int size: new int[] { 100, 500, 1000 } ) {
				int correct = 0;
				for( int attempt = 0; attempt < 10; attempt++ ) {
					System.err.println( "log2m: " + log2m + " size: " + size + " attempt: " + attempt );
					ArrayListMutableGraph mg = new ArrayListMutableGraph( size );
					for( int i = 1; i < size; i++ ) mg.addArc( 0, i );
					ImmutableGraph g = mg.immutableView();
					HyperApproximateNeighbourhoodFunction hyperApproximateNeighbourhoodFunction = new HyperApproximateNeighbourhoodFunction( g, attempt % 3 == 0 ? null : Transform.transpose( g ), log2m, null, 0, 10 * (attempt % 3), 10, attempt % 2 == 0, false, false, attempt );
					SequentialHyperApproximateNeighbourhoodFunction sequentialHyperApproximateNeighbourhoodFunction = new SequentialHyperApproximateNeighbourhoodFunction( g, log2m, null, attempt );
					hyperApproximateNeighbourhoodFunction.init();
					sequentialHyperApproximateNeighbourhoodFunction.init();
					final double current = hyperApproximateNeighbourhoodFunction.iterate();
					final double sequentialCurrent = sequentialHyperApproximateNeighbourhoodFunction.iterate();
					assertState( size, log2m, sequentialHyperApproximateNeighbourhoodFunction.registers(), hyperApproximateNeighbourhoodFunction.registers() );
					assertRelativeError( sequentialCurrent, current, THRESHOLD );
					if ( Math.abs( size * 2 - 1 - current ) <= 2 * rsd * ( size * 2 - 1 ) ) correct++;
					hyperApproximateNeighbourhoodFunction.close();
					sequentialHyperApproximateNeighbourhoodFunction.close();
				}
				assertTrue( size + ":" + rsd + " " + correct + " < " + 9, correct >= 9 );
			}
		}
	}

	@Test
	public void testTree() throws IOException {
		for( int log2m: new int[] { 4, 5, 6, 7, 8, 10, 12 } ) {
			double rsd = HyperApproximateNeighbourhoodFunction.relativeStandardDeviation( log2m );
			ImmutableGraph g = ArrayListMutableGraph.newCompleteBinaryIntree( 3 ).immutableView();
			final int[] correct = new int[ 3 ];
			for( int attempt = 0; attempt < 10; attempt++ ) {
				System.err.println( "log2m: " + log2m + " attempt: " + attempt );
				HyperApproximateNeighbourhoodFunction hyperApproximateNeighbourhoodFunction = new HyperApproximateNeighbourhoodFunction( g, attempt % 3 == 0 ? null : Transform.transpose( g ), log2m, null, 0, 10 * (attempt % 3), 10, attempt % 2 == 0, false, false, attempt );
				SequentialHyperApproximateNeighbourhoodFunction sequentialHyperApproximateNeighbourhoodFunction = new SequentialHyperApproximateNeighbourhoodFunction( g, log2m, null, attempt );
				hyperApproximateNeighbourhoodFunction.init();
				sequentialHyperApproximateNeighbourhoodFunction.init();

				if ( Math.abs( hyperApproximateNeighbourhoodFunction.iterate() - 29 ) <= 2 * rsd * 29 ) correct[ 0 ]++;
				sequentialHyperApproximateNeighbourhoodFunction.iterate();
				assertState( g.numNodes(), log2m, sequentialHyperApproximateNeighbourhoodFunction.registers(), hyperApproximateNeighbourhoodFunction.registers() );
				
				if ( Math.abs( hyperApproximateNeighbourhoodFunction.iterate() - 41 ) <= 2 * rsd * 41 ) correct[ 1 ]++;
				sequentialHyperApproximateNeighbourhoodFunction.iterate();
				assertState( g.numNodes(), log2m, sequentialHyperApproximateNeighbourhoodFunction.registers(), hyperApproximateNeighbourhoodFunction.registers() );
				
				if ( Math.abs( hyperApproximateNeighbourhoodFunction.iterate() - 49 ) <= 2 * rsd * 49 ) correct[ 2 ]++;
				sequentialHyperApproximateNeighbourhoodFunction.iterate();
				assertState( g.numNodes(), log2m, sequentialHyperApproximateNeighbourhoodFunction.registers(), hyperApproximateNeighbourhoodFunction.registers() );

				// Test that you can reuse the object
				
				hyperApproximateNeighbourhoodFunction.init();
				sequentialHyperApproximateNeighbourhoodFunction.init();

				if ( Math.abs( hyperApproximateNeighbourhoodFunction.iterate() - 29 ) <= 2 * rsd * 29 ) correct[ 0 ]++;
				sequentialHyperApproximateNeighbourhoodFunction.iterate();
				assertState( g.numNodes(), log2m, sequentialHyperApproximateNeighbourhoodFunction.registers(), hyperApproximateNeighbourhoodFunction.registers() );
				
				if ( Math.abs( hyperApproximateNeighbourhoodFunction.iterate() - 41 ) <= 2 * rsd * 41 ) correct[ 1 ]++;
				sequentialHyperApproximateNeighbourhoodFunction.iterate();
				assertState( g.numNodes(), log2m, sequentialHyperApproximateNeighbourhoodFunction.registers(), hyperApproximateNeighbourhoodFunction.registers() );
				
				if ( Math.abs( hyperApproximateNeighbourhoodFunction.iterate() - 49 ) <= 2 * rsd * 49 ) correct[ 2 ]++;
				sequentialHyperApproximateNeighbourhoodFunction.iterate();
				assertState( g.numNodes(), log2m, sequentialHyperApproximateNeighbourhoodFunction.registers(), hyperApproximateNeighbourhoodFunction.registers() );
				
				hyperApproximateNeighbourhoodFunction.close();
				sequentialHyperApproximateNeighbourhoodFunction.close();
			}
			//System.err.println( Arrays.toString( correct ));
			for( int i = 0; i < 3; i++ ) assertTrue( rsd + " " + correct[ i ] + " < " + 9, correct[ i ] >= 9 );
		}
	}
	
	@Test(expected=IllegalStateException.class)
	public void testInitClosed() throws IOException {
		ImmutableGraph g = ArrayListMutableGraph.newCompleteBinaryIntree( 3 ).immutableView();
		HyperApproximateNeighbourhoodFunction hyperApproximateNeighbourhoodFunction = new HyperApproximateNeighbourhoodFunction( g, 8 );
		hyperApproximateNeighbourhoodFunction.close();
		hyperApproximateNeighbourhoodFunction.init();
	}

	@Test(expected=IllegalStateException.class)
	public void testInitIterate() throws IOException {
		ImmutableGraph g = ArrayListMutableGraph.newCompleteBinaryIntree( 3 ).immutableView();
		HyperApproximateNeighbourhoodFunction hyperApproximateNeighbourhoodFunction = new HyperApproximateNeighbourhoodFunction( g, 8 );
		hyperApproximateNeighbourhoodFunction.close();
		hyperApproximateNeighbourhoodFunction.iterate();
	}
}
